/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.*;
import java.nio.file.Files;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

abstract class OutputStreamFactory implements AutoCloseable {
    static class DirOutputStreamFactory extends OutputStreamFactory {
        protected final File destDir;

        DirOutputStreamFactory(File destDir) {
            this.destDir = destDir;
        }
        String getFilename(String filename) {
            return (destDir==null || new File(filename).isAbsolute())?filename:new File(destDir,filename).getPath();
        }
        OutputStream newOutputStream(String filename) throws IOException {
            //System.out.println("newOutputStream for : "+filename);
            File outfile = (destDir==null || new File(filename).isAbsolute())?new File(filename).getAbsoluteFile():new File(destDir,filename).getAbsoluteFile();
            File parentfile = outfile.getParentFile();
            if (!parentfile.exists()) {
                if (!parentfile.mkdirs())
                    throw new IOException("Failed to create directory <" + parentfile.getPath() + ">.");
            }
            return new FileOutputStream(outfile);
        }
    }

    static class TempOutputStreamFactory extends DirOutputStreamFactory {
        private final File tempDir;

        TempOutputStreamFactory(File destDir, String tempDir) {
            super(destDir);
            this.tempDir = new File(tempDir).getAbsoluteFile();
            if (!this.tempDir.exists())
                this.tempDir.mkdirs();
        }

        OutputStream newOutputStream(String filename) throws IOException {
            File tempFile = new File(filename).isAbsolute()?File.createTempFile(new File(filename).getName(),".tmp",tempDir).getAbsoluteFile():new File(tempDir,filename).getAbsoluteFile();
            File destFile = (destDir==null || new File(filename).isAbsolute())?new File(filename).getAbsoluteFile():new File(destDir,filename).getAbsoluteFile();
            Files.createDirectories(tempFile.toPath().getParent());
            return new RenameableFileOutputStream(new FileOutputStream(tempFile), tempFile.toPath(), destFile.toPath());
        }
    }

    static class ZipEntryOutputStreamFactory extends OutputStreamFactory {
        private final ZipOutputStream zipOutputStream;

        ZipEntryOutputStreamFactory(ZipOutputStream zipOutputStream) {
            this.zipOutputStream = zipOutputStream;
        }

        OutputStream newOutputStream(String filename) throws IOException {
            zipOutputStream.putNextEntry(new ZipEntry(filename));
            return new ZipEntryOutputStream(zipOutputStream);
        }

        @Override
        public void close() throws IOException {
            zipOutputStream.close();
        }
        void tryParallelStream(AbstractTool task) throws IOException {
            throw new IOException(task.getClass().getSimpleName()+":"+task.taskname + " tries to open a second Outputfile while first is not closed. Outputarchive zip doesn't support writing in parallel.");
        }

    }

    abstract OutputStream newOutputStream(String filename) throws IOException;
    String getFilename(String filename) {
        return filename;
    }
    void tryParallelStream(AbstractTool task) throws IOException {
    }
    @Override
    public void close() throws IOException {

    }
}