/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.*;
import java.nio.file.Files;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.zip.ZipOutputStream;

import com.enitsys.sqltools.CSVFormat.CSVDefinition;

public final class SQL2csv extends AbstractSQLTool implements Archivable {
	private static final String formatWithToString="toString";
	private static final String formatSignificant="significant";
	public String significant(Object value) {
		return significant.format(value);
	}

	private final String stmtString;
	private final DecimalFormat significant;
	private final CSVDefinition csvDefinition;

	SQL2csv(String taskname, String connection, String directory)throws IOException {
		super( taskname, connection, directory);
		CSVFormat csvFormat = new CSVFormat();
		csvFormat.loadFormatDefinitions(configDir, taskProperties);
		this.csvDefinition=CSVDefinition.valueOf(csvFormat, taskProperties);
		this.stmtString=stmt(null, taskname + ".sql");
		DecimalFormat df=new DecimalFormat("0.0", DecimalFormatSymbols.getInstance(Locale.US));
		df.setMaximumFractionDigits(400);
		df.setMinimumFractionDigits(1);
		df.setMinimumIntegerDigits(1);
		df.setMaximumIntegerDigits(400);
		significant=df;
	}
	private String[] getTemplates(ResultSetMetaData md, BaseFormat[] formats) throws SQLException {
		String[] result=new String[formats.length];
		for(int col=0;col<result.length;col++) {
			result[col]=formats[col].getFormat(taskProperties, md.getColumnLabel(col+1));
			if (formatWithToString.equals(result[col]))
				result[col]=null;
		}
		return result;
	}
	private void encode(ResultSet resultSet, PrintWriter out) throws SQLException {
		long start=System.currentTimeMillis();
		ResultSetMetaData md = resultSet.getMetaData();
		final int colcount = md.getColumnCount();
		BaseFormat[] formats=getFormats(md);
		String[] templates=getTemplates(md, formats);
		int buffersize=formats.length * 10;
		StringBuilder line=new StringBuilder(buffersize);

		if (csvDefinition.header()) {
			for (int col = 1; col <= colcount; col++) {
				if (col > 1)
					line.append(csvDefinition.separator());
				if (csvDefinition.quoteHeader())
					line.append(csvDefinition.quote(md.getColumnLabel(col)));
				else
					line.append(md.getColumnLabel(col));
			}
			out.println(line);
		}
		LOG.fine("SQL-2-meta      : "+(System.currentTimeMillis()-start)+ " millis.");
		start=System.currentTimeMillis();
		Object value;
		Locale locale=Locale.US;
		String strValue;
		int separators;
		long linenr=0L;
		long millis_read=0L;
		long millis_format=0L;
		long millis_fetch=0L;
		long millis_write=0L;
		Object[] values = new Object[formats.length];
		while(resultSet.next()) {
			if (linenr==0)
				LOG.fine("SQL-3-read-first: "+(System.currentTimeMillis()-start)+ " millis.");
			else
				millis_read += (System.currentTimeMillis()-start);
			start=System.currentTimeMillis();
			linenr++;

			line=new StringBuilder(buffersize);
			separators = 0;
			//values = new Object[formats.length];
			for(int col=0;col<formats.length;col++) {
				values[col] = formats[col].getValue(resultSet, col + 1);
			}
			millis_fetch += (System.currentTimeMillis()-start);
			start=System.currentTimeMillis();
			for(int col=0;col<formats.length;col++) {
				value=values[col];

				if (col > 0)
					separators++;
				if (value == null)
					continue;
				while (separators>0) {
					line.append(csvDefinition.separator());
					separators--;
				}
				if (templates[col]==null)
					strValue=value.toString();
				else if (formatSignificant.equals(templates[col]))
					strValue=significant(value);
				else
					strValue=String.format(locale, templates[col] , value);
				line.append(csvDefinition.encode(strValue, formats[col]));
			}
			if (csvDefinition.rectangle() && (line.length()>0 || !csvDefinition.suppressEmptyLines()))
				while (separators>0) {
					line.append(csvDefinition.separator());
					separators--;
				}
			millis_format += (System.currentTimeMillis()-start);
			start=System.currentTimeMillis();

			if (line.length() == 0 && csvDefinition.suppressEmptyLines())
				continue;

			if (buffersize < line.length())
				buffersize = line.length();

			out.println(line);

			millis_write += (System.currentTimeMillis()-start);
			start=System.currentTimeMillis();
		}
		out.flush();
		millis_write += (System.currentTimeMillis()-start);
		LOG.fine("SQL-4-read      : "+millis_read+ " millis.");
		LOG.fine("SQL-5-fetch     : "+millis_fetch+ " millis.");
		LOG.fine("SQL-6-format    : "+millis_format+ " millis.");
		LOG.fine("SQL-7-write     : "+millis_write+ " millis.");
		LOG.fine(linenr+" lines written.");

	}
	private void write(Connection dataConnection, OutputStream outputStream) throws SQLException, IOException {
		if (dataConnection == null) {
			try (Connection conn = createConnection()) {
				write(conn, outputStream);
			}
			return;
		}

		long start=System.currentTimeMillis();

		try (PrintWriter out = new ModifiablePrintWriter(new OutputStreamWriter(outputStream, taskProperties.getProperty("out.encoding", "UTF-8")), null)) {
			try (Statement statement=dataConnection.createStatement()) {
				System.setProperty("timestamp.execute", new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()));
				String fetchsize=taskProperties.getProperty("sql.fetchsize");
				if (fetchsize!=null)
					statement.setFetchSize(Integer.parseInt(fetchsize));
				try (ResultSet resultSet=statement.executeQuery(stmtString)) {
					LOG.fine("SQL-1-execute   : "+(System.currentTimeMillis()-start)+ " millis.");

					encode(resultSet, out);
				}
			}
		}
		LOG.fine("File generated in "+(System.currentTimeMillis()-start)+ " millis.");
	}

	private void write(Connection dataConnection, OutputStreamFactory outputStreamFactory, String filename, boolean addSQLFile) throws SQLException, IOException {

		if (addSQLFile) {
			try (PrintWriter printWriter = new ModifiablePrintWriter(new OutputStreamWriter(outputStreamFactory.newOutputStream(filename + ".sql"), taskProperties.getProperty("out.encoding", "UTF-8")), null)) {
				printWriter.println(stmtString);
				printWriter.flush();
			}
		}
		write(dataConnection, outputStreamFactory.newOutputStream(filename));
	}

	public void write(Connection dataConnection, OutputStreamFactory outputStreamFactory, boolean addSQLFile) throws SQLException, IOException {
		String filename = replace(taskProperties.getProperty("out.file"), replacements);
		write(dataConnection,outputStreamFactory,filename, addSQLFile);
	}
	public void run() throws SQLException, IOException {
		try (Connection conn = createConnection()) {
			String zipFile = replace(System.getProperty("out.zip.file", taskProperties.getProperty("out.zip.file")), replacements);
			if (zipFile != null) {
				File outfile = (destDir == null || new File(zipFile).isAbsolute()) ? new File(zipFile) : new File(destDir, zipFile);
				Files.createDirectories(outfile.toPath().getParent());
				try (OutputStreamFactory outputStreamFactory = new OutputStreamFactory.ZipEntryOutputStreamFactory(new ZipOutputStream(new FileOutputStream(outfile)))) {
					write(conn, outputStreamFactory, false);
				}
			} else {
				String tempDir = System.getProperty("out.tempdir", taskProperties.getProperty("out.tempdir"));
				OutputStreamFactory outputStreamFactory = null;
				if (tempDir != null)
					outputStreamFactory = new OutputStreamFactory.TempOutputStreamFactory(destDir, tempDir);
				else
					outputStreamFactory = new OutputStreamFactory.DirOutputStreamFactory(destDir);
				write(conn, outputStreamFactory, false);
			}
		}
	}
	public static void main(String[] args) {
		main(SQL2csv.class, args);

	}
}
