/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipOutputStream;

public final class SQL2xml extends AbstractSQLTool implements Archivable {

    class ElementDescriptor {
        class Attr {
            String xmlName;
            String sqlName;
            int sqlCol;
            String xmlType;
            private Attr(String xmlName, int sqlCol, String xmlType) {
                this.xmlName = xmlName;
                this.sqlName = null;
                this.sqlCol = sqlCol;
                this.xmlType = xmlType==null ? "A" : xmlType;
            }
            private Attr(String xmlName, String sqlName, String xmlType) {
                this.xmlName = xmlName;
                this.sqlName = sqlName;
                this.sqlCol = 0;
                this.xmlType = xmlType==null ? "A" : xmlType;
            }

        }
        final ElementDescriptor parent;
        final String domainPrefix;
        final String domain;
        final String name;
        final String sql;
        final List<ElementDescriptor> elements=new ArrayList<>();
        final List<ElementDescriptor> sqlElements=new ArrayList<>();
        final List<String[]> attributes = new ArrayList<>();
        final List<Attr> sqlAttributes = new ArrayList<>();
        final Map<String, Attr> sqlAttrByName = new HashMap<String, SQL2xml.ElementDescriptor.Attr>();
        final Map<Integer, Attr> sqlAttrById = new HashMap<Integer, SQL2xml.ElementDescriptor.Attr>();
        final List<String> sqlKeys = new ArrayList<>();
        final List<String[]> filter = new ArrayList<>();
        final String dynElement;
        final Attr dynSql;

        final boolean allSqlAttributes;
        private boolean openHeader=false;
        private String currentKey=null;
        private boolean parentKey=true;
        private boolean inDyn=false;
        ElementDescriptor(ElementDescriptor parent, String domainPrefix, String domain) throws IOException {
            this.parent = parent;
            this.domainPrefix = domainPrefix;
            this.domain = domain;
            this.name=taskProperties.getProperty(domainPrefix+domain+".name", domain);
            for (StringTokenizer st=new StringTokenizer(taskProperties.getProperty(domainPrefix+domain+".attributes",""),",;");st.hasMoreTokens();) {
                String[] attr=new String[3];
                attr[1]=st.nextToken().trim();
                if (attr[1] != null && attr[1].toUpperCase().endsWith("/T")) {
                    attr[0] = attr[1].substring(5);
                    attr[1]=encodeText(taskProperties.getProperty(domainPrefix+domain+".attribute."+attr[0],"").trim());
                    attr[2] = "T";
                } else if (attr[1] != null && attr[1].toUpperCase().endsWith("/C")) {
                    attr[0] = attr[1].substring(5);
                    attr[1]=encodeCData(taskProperties.getProperty(domainPrefix+domain+".attribute."+attr[0],"").trim());
                    attr[2] = "C";
                } else {
                    attr[0] = attr[1];
                    attr[1]=encodeValue(taskProperties.getProperty(domainPrefix+domain+".attribute."+attr[0],"").trim());
                    attr[2] = "A";
                }
                if (!(attr[0]==null || attr[0].isEmpty() || attr[1].isEmpty()))
                    this.attributes.add(attr);
            }
            String sql = stmt(domainPrefix+domain, null);
            this.sql=(sql==null?null:replaceSQL(sql, replacements));
            for (StringTokenizer st=new StringTokenizer(taskProperties.getProperty(domainPrefix+domain+".elements",""),",;");st.hasMoreTokens();) {
                String elementDomain=st.nextToken().trim();
                if (!elementDomain.isEmpty())
                    elements.add(new ElementDescriptor(this, "element.",elementDomain));
            }
            for (StringTokenizer st=new StringTokenizer(taskProperties.getProperty(domainPrefix+domain+".sql.elements",""),",;");st.hasMoreTokens();) {
                String sqlElementDomain=st.nextToken().trim();
                if (!sqlElementDomain.isEmpty())
                    sqlElements.add(new ElementDescriptor(this, "element.",sqlElementDomain));
            }
            String sqlAttributes=taskProperties.getProperty(domainPrefix+domain+".sql.attributes");
            Attr dynSql =null;
            if (sqlAttributes!=null) {
                sqlAttributes=sqlAttributes.trim();
                if (!sqlAttributes.isEmpty()) {
                    this.allSqlAttributes = sqlAttributes.startsWith("*");
                    for (StringTokenizer st=new StringTokenizer(sqlAttributes,",;");st.hasMoreTokens();) {
                        String attrStr = st.nextToken().trim();
                        if (attrStr.startsWith("*")) {
                            //TODO
                        } else {
                            Matcher m=sqlAttr1.matcher(attrStr);
                            if (m.matches()) {
                                Attr a=new Attr(m.group(3), Integer.parseInt(m.group(1)), m.group(5));
                                if ("S".equals(a.xmlType))
                                    dynSql=a;
                                else {
                                    this.sqlAttributes.add(a);
                                    this.sqlAttrById.put(a.sqlCol, a);
                                }
                            } else {
                                m=sqlAttr2.matcher(attrStr);
                                if (m.matches()) {
                                    Attr a=new Attr(m.group(3), m.group(1), m.group(5));
                                    if ("S".equals(a.xmlType))
                                        dynSql=a;
                                    else {
                                        this.sqlAttributes.add(a);
                                        this.sqlAttrByName.put(a.sqlName, a);
                                    }
                                }
                            }
                        }
                    }
                } else
                    allSqlAttributes=false;
            } else
                allSqlAttributes=false;
            this.dynSql=dynSql;
            if (dynSql!=null)  {
                this.dynElement=taskProperties.getProperty(domainPrefix+domain+".sql.dyn-element");
                //System.out.println("dynSql found, dynElement="+this.dynElement);
            }
            else
                this.dynElement=null;
            String sqlKeys=taskProperties.getProperty(domainPrefix+domain+".sql.key");
            if (sqlKeys!=null) {
                sqlKeys=sqlKeys.trim();
                if (!sqlKeys.isEmpty()) {
                    for (StringTokenizer st=new StringTokenizer(sqlKeys,",;");st.hasMoreTokens();) {
                        String key = st.nextToken().trim();
                        if (!key.isEmpty())
                            this.sqlKeys.add(key);
                    }
                }
            }
            String sqlFilter=taskProperties.getProperty(domainPrefix+domain+".sql.filter");
            if (sqlFilter!=null) {
                sqlFilter=sqlFilter.trim();
                if (!sqlFilter.isEmpty()) {
                    for (StringTokenizer st=new StringTokenizer(sqlFilter,",;");st.hasMoreTokens();) {
                        String[] filter = new String[2];
                        filter[0] = st.nextToken().trim();
                        if (!filter[0].isEmpty()) {
                            filter[1] = taskProperties.getProperty(domainPrefix+domain+".sql.filter."+filter[0],"").trim();
                            if (!filter[0].isEmpty())
                                this.filter.add(filter);
                        }
                    }
                }
            }
        }
        void writeAttributes(PrintWriter out, List<String[]> textElements) {
            for(String[] attr:attributes) {
                if ("T".equals(attr[2]))
                    textElements.add(attr);
                else
                    out.print(arg(attr[0],attr[1]));
            }
        }

        void writeAttribute(PrintWriter out, ResultSet rs, ResultSetMetaData rsmd, Attr a, List<String[]> textElements) throws SQLException {
            if (a.xmlName == null) {
                if (a.sqlName!=null)
                    a.xmlName = toXmlName(a.sqlName);
                else
                    a.xmlName = toXmlName(rsmd.getColumnName(a.sqlCol));
            }

            if ("X".equals(a.xmlType))
                return;
            else if ("T".equals(a.xmlType)) {
                String[] attr=new String[3];
                attr[0] = a.xmlName;
                attr[2] = a.xmlType;
                attr[1] = encodeText(a.sqlCol>0?rs.getString(a.sqlCol):rs.getString(a.sqlName));
                if (attr[0] !=null && attr[1]!=null)
                    textElements.add(attr);
            }
            else if ("C".equals(a.xmlType)) {
                String[] attr=new String[3];
                attr[0] = a.xmlName;
                attr[2] = a.xmlType;
                attr[1] = encodeCData(a.sqlCol>0?rs.getString(a.sqlCol):rs.getString(a.sqlName));
                if (attr[0] !=null && attr[1]!=null)
                    textElements.add(attr);
            }
            else if (a.sqlCol>0)
                out.print(arg(rs,rsmd,a.sqlCol,a.xmlName));
            else
                out.print(arg(rs,rsmd,a.sqlName,a.xmlName));



        }
        void writeAttributes(PrintWriter out, ResultSet rs, ResultSetMetaData rsmd, List<String[]> textElements) throws SQLException {
            if (allSqlAttributes) {
                int cols=rsmd.getColumnCount();
                for (int i=1;i<=cols;i++) {
                    Attr a = this.sqlAttrById.get(i);
                    if (a==null)
                        a = this.sqlAttrByName.get(rsmd.getColumnName(i));
                    if(a==null)
                        out.print(arg(rs, rsmd, i, null));
                    else
                        writeAttribute(out, rs, rsmd, a, textElements);
                }
            } else
                for (Attr a:sqlAttributes)
                    writeAttribute(out, rs, rsmd, a, textElements);

        }
        void closeHeader(PrintWriter out) {
            if (parent!=null)
                parent.closeHeader(out);
            if(openHeader) {
                out.println(">");
                openHeader=false;
            }
        }
        private void addKey(ResultSet rs, ResultSetMetaData rsMD, StringBuilder buffer, Attr col) throws SQLException {
            buffer.append("|[");
            String value = col.sqlCol>0 ? rs.getString(col.sqlCol):rs.getString(col.sqlName);
            if (value!=null && !value.isEmpty())
                buffer.append(value);
            buffer.append("]");
        }
        private void addKey(ResultSet rs, ResultSetMetaData rsMD, StringBuilder buffer, String col) throws SQLException {
            buffer.append("|[");
            String value = rs.getString(col);
            if (value!=null && !value.isEmpty())
                buffer.append(value);
            buffer.append("]");
        }

        String key(ResultSet rs, ResultSetMetaData rsmd) throws SQLException {
            StringBuilder sb;
            if (sqlKeys.isEmpty()) {
                sb=new StringBuilder(sqlAttributes.size()*50);
                for (Attr attr:sqlAttributes)
                    addKey(rs, rsmd, sb, attr);
            } else {
                sb=new StringBuilder(sqlKeys.size()*50);
                for (String attr:sqlKeys)
                    addKey(rs, rsmd, sb, attr);
            }
            return sb.toString();
        }
        boolean isKey(ResultSet rs, ResultSetMetaData rsmd) throws SQLException {
            if (inDyn)
                return true;
            if (currentKey==null)
                return true;
            if (!currentKey.equals(key(rs,rsmd)))
                return false;
            if (parentKey && parent!=null && !parent.isKey(rs, rsmd))
                return false;
            return true;
        }
        boolean isFilter(ResultSet rs, ResultSetMetaData rsmd) throws SQLException {
            if (this.sql!=null || rs==null || inDyn)
                return true;
            if (parent!=null && ! parent.isFilter(rs, rsmd))
                return false;
            for (String[] filter:this.filter) {
                String val=rs.getString(filter[0]);
                if (!filter[1].equals(val))
                    return false;
            }
            return true;
        }
        boolean write(PrintWriter out, String prefix, ResultSet parentRS, ResultSetMetaData parentRSMD,boolean[] consumed) throws SQLException, IOException {
            Connection conn=null;
            Statement stmt=null;
            ResultSet rs=null;
            ResultSetMetaData rsmd=null;
            boolean useCurrentRow=true;
            currentKey=null;
            try {
                if (sql!=null) {
                    conn=createConnection(logConnection);
                    logConnection=false;
                    stmt=conn.createStatement();
                    LOG.fine("executing [[["+sql+"]]]");
                    rs = stmt.executeQuery(sql);
                    rsmd = rs.getMetaData();
                    useCurrentRow=rs.next();
                    parentKey=false;
                } else {
                    rs=parentRS;
                    rsmd=parentRSMD;
                    parentKey=true;
                    useCurrentRow=(rs!=null)&&(!rs.isAfterLast());
                }
                if (!isFilter(useCurrentRow?rs:null, rsmd))
                    return false;
                if (name!=null) {
                    List<String[]> textElements = new ArrayList<>();
                    closeHeader(out);
                    out.print(prefix+"<"+xmlname(name));
                    openHeader=true;
                    writeAttributes(out, textElements);
                    if (useCurrentRow)
                        writeAttributes(out, rs, rsmd, textElements);
                    if (!textElements.isEmpty()) {
                        closeHeader(out);
                        for(String[] textElement:textElements) {
                            out.println(prefix+TAB+"<"+xmlname(textElement[0])+">"
                                    +("C".equals(textElement[2])?"<![CDATA[":"")
                                    +textElement[1]
                                    +("C".equals(textElement[2])?"]]>":"")
                                    +"</"+xmlname(textElement[0])+">");
                        }
                    }
                }
                for (ElementDescriptor element:elements)
                    element.write(out, prefix+TAB, useCurrentRow?rs:null, rsmd,consumed);

                if (rs!=null && useCurrentRow && !sqlElements.isEmpty()) {
                    currentKey=key(rs,rsmd);
                    while (useCurrentRow) {
                        consumed[0] = true;
                        for (ElementDescriptor element:sqlElements)
                            if (element.write(out, prefix+TAB, rs, rsmd, consumed))
                                break;
                        if (consumed[0])
                            useCurrentRow = rs.next();
                        else
                            useCurrentRow = !rs.isAfterLast();
                        if (!useCurrentRow) {
                            consumed[0] = false;
                        } else if (!isKey(rs, rsmd)) {
                            consumed[0] = false;
                            useCurrentRow = false;
                        } else if (!isFilter(rs, rsmd)) {
                            consumed[0] = false;
                            useCurrentRow = false;
                        }
                    }
                }

                if (dynElement!=null && !dynElement.isEmpty()) {
                    ElementDescriptor dynElement=new ElementDescriptor(this, "element.", this.dynElement);
                    String sqlDyn =null;
                    if (dynSql.sqlCol>0 && rsmd.getColumnCount()>=dynSql.sqlCol)
                        sqlDyn=rs.getString(dynSql.sqlCol);
                    else if (dynSql.sqlCol<=0)
                        sqlDyn=rs.getString(dynSql.sqlName);
                    if (sqlDyn!=null) {
                            try (Connection connDyn=createConnection(logConnection)) {
                                logConnection=false;
                                try (Statement stmtDyn = connDyn.createStatement()) {
                                    LOG.fine("executing [[[" + sqlDyn + "]]]");
                                    try (ResultSet rsDyn = stmtDyn.executeQuery(sqlDyn)) {
                                        ResultSetMetaData rsmdDyn = rsDyn.getMetaData();
                                        boolean[] consDyn = {false};
                                        boolean moreRows = rsDyn.next();
                                        inDyn = true;
                                        while (moreRows) {
                                            consDyn[0] = true;
                                            dynElement.write(out, prefix + TAB, rsDyn, rsmdDyn, consDyn);
                                            if (consDyn[0])
                                                moreRows = rsDyn.next();
                                            else
                                                moreRows = !rsDyn.isAfterLast();
                                        }
                                        inDyn = false;
                                    }
                                }
                            }
                    }

                }
                if(name!=null) {
                    if (openHeader)
                        out.println("/>");
                    else
                        out.println(prefix+"</"+xmlname(name)+">");
                    openHeader=false;
                }
                return true;
            } finally {
                if (sql !=null) {
                    if (rs !=null)
                        try {
                            rs.close();
                        } catch (Exception e) {

                        }
                    if (stmt != null)
                        try {
                            stmt.close();
                        } catch (Exception e) {

                        }
                    if (conn != null)
                        try {
                            conn.close();
                        } catch (Exception e) {

                        }
                }
            }
        }
    }

    class RootDescriptor extends ElementDescriptor {
        RootDescriptor() throws IOException {
            super(null, "", "root");
        }
        @Override
        void writeAttributes(PrintWriter out, List<String[]> textElements) {
            if (xmlNS == null)
                out.print(arg("xmlns",nsURI));
            else
                out.print(arg("xmlns:"+xmlNS,nsURI));
            super.writeAttributes(out, textElements);
        }
        @Override
        boolean write(PrintWriter out, String prefix, ResultSet parentRS, ResultSetMetaData parentRSMD, boolean[] consumed) throws SQLException, IOException {
            out.println("<?xml version=\"1.0\" encoding=\""+taskProperties.getProperty("xml.encoding",taskProperties.getProperty("out.encoding", "UTF-8"))+"\"?>");
            return super.write(out, prefix, parentRS, parentRSMD, consumed);
        }
    }
    /*
    final Properties connectionProperties;
    final Properties taskProperties=new Properties();
     */
    final String TAB = "    ";
    /*
        final File configDir;
        final Properties replacements;
        final String taskname;
     */
    final String xmlNS;
    final String nsURI;
    final RootDescriptor root;
    private boolean logConnection = true;

    //	static final Pattern replace=Pattern.compile("\\[([^\\[\\]]+)\\]");
//	static final Pattern replaceSQL=Pattern.compile("\\[\\[([^\\[\\]]+)\\]\\]");
    static final Pattern sqlAttr1=Pattern.compile("^([0-9]+)(/([^/]+)?)?(/([ACSTX]))?$");
    static final Pattern sqlAttr2=Pattern.compile("^([^/]+)(/([^/]+)?)?(/([ACSTX]))?$");
    /*
        static  String replace(String stmt, Properties replacements) {
            StringBuffer result = new StringBuffer(stmt.length());
            Matcher matcher=replace.matcher(stmt);
            while (matcher.find())
                matcher.appendReplacement(result, replacements.getProperty(matcher.group(1)));
            matcher.appendTail(result);
            return result.toString();
        }
        static  String replaceSQL(String stmt, Properties replacements) {
            StringBuffer result = new StringBuffer(stmt.length());
            Matcher matcher=replaceSQL.matcher(stmt);
            while (matcher.find())
                matcher.appendReplacement(result, replacements.getProperty(matcher.group(1),""));
            matcher.appendTail(result);
            return result.toString();
        }

     */
    static String encodeText(String text) {
        if (text==null)
            return null;
        boolean foundUndefined=false;
		/*
		for (int cp:text.codePoints().toArray()) 
			if (!Character.isDefined(cp)) {
				foundUndefined=true;
				System.out.println("undefined unicode codepoint: "+cp);
			}
		*/
        if (foundUndefined)
            LOG.warning("undefined unicode codepoint(s) found in: "+text);

        return text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace((char)0, ' ').replace(String.valueOf((char)133), "...");
    }
    static String encodeValue(String text) {
        if (text==null)
            return null;
        return text.replace("&", "&amp;").replace("<", "&lt;").replace("\"", "&quot;").replace((char)0, ' ');
    }
    static String encodeCData(String text) {
        if (text==null)
            return null;
        return text.replace("]]>", "]]]]><![CDATA[>").replace((char)0, ' ');
    }
    String xmlname(String simpleName) {
        return ((xmlNS==null||xmlNS.isEmpty()||simpleName.contains(":"))?simpleName:(xmlNS+":"+simpleName));
    }
    SQL2xml(String taskname, String connection, String directory)throws IOException {
        super( taskname,connection, directory);
		/*
		this.connectionProperties=connectionProperties;
		this.taskProperties.load(new FileInputStream(new File(configDir, taskname+".properties")));
		this.configDir=configDir;
		this.replacements=new Properties(taskProperties);
		this.replacements.putAll(replacements);
		this.taskname=taskname;
		*/

        this.xmlNS =  taskProperties.getProperty("xmlns");
        this.nsURI =  taskProperties.getProperty("nsuri", this.xmlNS==null?null:"test");
        this.root = new RootDescriptor();
		/*
		System.out.println("---------- stmt --------");
		System.out.println(this.stmtString);
		System.out.println("---------- stmt --------");
		*/
    }

    String arg(String name, String value) {
        if (value==null || value=="")
            return "";
        return " " + xmlname(name) + "=\"" + value + "\"";
    }
    String toXmlName(String sqlName) {
        return sqlName.toLowerCase().replace("_","-");
    }
    String arg(ResultSet rs, ResultSetMetaData rsMD, int col, String attrName) throws SQLException {
        if (rsMD.getColumnCount()<col)
            return "";
        String value = rs.getString(col);
        if (value==null || value.isEmpty())
            return "";
        return " " + xmlname(attrName==null?toXmlName(rsMD.getColumnName(col)):attrName) + "=\"" + encodeValue(value) + "\"";
    }
    String arg(ResultSet rs, ResultSetMetaData rsMD, String col, String attrName) throws SQLException {
        try {
            String value = rs.getString(col);
            if (value==null || value.isEmpty())
                return "";
            return " " + xmlname(attrName==null?toXmlName(col):attrName) + "=\"" + encodeValue(value) + "\"";
        }catch(SQLException e) {
            throw new SQLException("Error in getString(\""+col+"\")", e);
        }
    }
    public void write(Connection dataConnection, OutputStreamFactory outputStreamFactory, boolean addSQLFile) throws SQLException, IOException {
        String filename=replace(taskProperties.getProperty("out.file"), replacements);
        try (PrintWriter printWriter = new ModifiablePrintWriter(new OutputStreamWriter(outputStreamFactory.newOutputStream(filename), taskProperties.getProperty("out.encoding", "UTF-8")), null)) {
            root.write(printWriter, "", null, null, new boolean[]{true});
        }
    }
    public void run() throws SQLException, IOException{
        String zipFile = replace(System.getProperty("out.zip.file", taskProperties.getProperty("out.zip.file")), replacements);
        if (zipFile != null) {
            File outfile = (destDir == null || new File(zipFile).isAbsolute()) ? new File(zipFile) : new File(destDir, zipFile);
            Files.createDirectories(outfile.toPath().getParent());
            try (OutputStreamFactory outputStreamFactory = new OutputStreamFactory.ZipEntryOutputStreamFactory(new ZipOutputStream(new FileOutputStream(outfile)))) {
                write(null, outputStreamFactory, false);
            }
        } else {
            String tempDir = System.getProperty("out.tempdir", taskProperties.getProperty("out.tempdir"));
            OutputStreamFactory outputStreamFactory = null;
            if (tempDir != null && !tempDir.trim().isEmpty())
                outputStreamFactory = new OutputStreamFactory.TempOutputStreamFactory(destDir, tempDir);
            else
                outputStreamFactory = new OutputStreamFactory.DirOutputStreamFactory(destDir);
            write(null, outputStreamFactory, false);
        }
    }
    public static void main(String[] args) {
/*
        System.setProperty("logging.level", "FINE");
        System.setProperty("logging.simple", "true");
        System.setProperty("configuration.dir", "src/examples/config");
        System.setProperty("sql.dir", "src/examples/sql");
        System.setProperty("out.dir", "test/out");
        System.setProperty("connection.dir", "jdbcconnection");

        String[] xargs=(args==null || args.length == 0)?new String[]{"sql2xml.dbmeta.Oracle","exampleconnection"}:args;

        main(SQL2xml.class, xargs);
*/
        main(SQL2xml.class, args);
    }
}
