/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import com.enitsys.sqltools.CSVFormat.CSVDefinition;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.*;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

public final class SQLAnalyser extends AbstractSQLTool {
	private static final String formatWithToString="toString";
	private static final String formatSignificant="significant";

	private static final char[] hexChar = "0123456789ABCDEF".toCharArray();
	private static final String toHex(byte[] bytes) {
		StringBuilder result=new StringBuilder(bytes.length*2);
		for(byte b:bytes)
			result.append(hexChar[(b >> 4) & 0xF]).append(hexChar[b & 0xF]);
		return result.toString();
	}

	public String significant(Object value) {
		return significant.format(value);
	}


	private final String stmtString;
	private final DecimalFormat significant;
//	private final CSVFormat csvFormat = new CSVFormat();
//	private final CSVDefinition csvDefinition;

	SQLAnalyser(String taskname, String connection, String directory)throws IOException {
		super( taskname, connection, directory);
//		csvFormat.loadFormatDefinitions(configDir, taskProperties);
//		this.csvDefinition=CSVDefinition.valueOf(csvFormat, taskProperties);
		this.stmtString=stmt(null, taskname + ".sql");
		DecimalFormat df=new DecimalFormat("0.0", DecimalFormatSymbols.getInstance(Locale.US));
		df.setMaximumFractionDigits(400);
		df.setMinimumFractionDigits(1);
		df.setMinimumIntegerDigits(1);
		df.setMaximumIntegerDigits(400);
		significant=df;
	}
	public void run() throws SQLException, IOException{
		try (Connection conn =createConnection()) {
			try (Statement stmt = conn.createStatement()) {
				System.setProperty("timestamp.execute", new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()));
				String fetchsize=taskProperties.getProperty("sql.fetchsize");
				if (fetchsize!=null)
					stmt.setFetchSize(Integer.parseInt(fetchsize));
				try (ResultSet rs = stmt.executeQuery(stmtString)) {
					ResultSetMetaData md = rs.getMetaData();
					final int colcount = md.getColumnCount();
					BaseFormat[] formats=getFormats(md);
					byte[][] currentHashes=new byte[colcount+1][];
					byte[][] consolidatedHashes=new byte[colcount+1][];
					String tmpDirStr=System.getProperty("out.tempdir",taskProperties.getProperty("out.tempdir"));
					File tmpDir=null;
					if(tmpDirStr!=null) {
						tmpDir=new File(tmpDirStr).getAbsoluteFile();
						if(!tmpDir.exists())
							tmpDir.mkdirs();
					}
					/*
					File destDir=new File(taskProperties.getProperty("directory"));
					if (!destDir.exists())
						destDir.mkdirs();
					 */
					String filename=replace(taskProperties.getProperty("out.file",taskname+".txt"), System.getProperties());
					File outfile=new File(tmpDir==null?destDir:tmpDir,filename);
					PrintWriter out=new PrintWriter(new OutputStreamWriter(
							new FileOutputStream(outfile), taskProperties.getProperty("out.encoding", "UTF-8"))); 
//					StringBuilder line=new StringBuilder();
					MessageDigest colMessageDigest;
					MessageDigest lineMessageDigest;
					try {
						colMessageDigest = MessageDigest.getInstance("MD5");
						lineMessageDigest = MessageDigest.getInstance("MD5");
					} catch (NoSuchAlgorithmException e) {
						throw new RuntimeException(e);
					}
					final byte[] empty = colMessageDigest.digest();
					Object value;
					Locale locale=Locale.US;
					long rows=0L;
					long[] nulls=new long[colcount];
					while(rs.next()) {
						rows++;
						for(int col=0;col<formats.length;col++) {
							value=formats[col].getValue(rs, col+1);
							if(value!=null)
								currentHashes[col + 1] = colMessageDigest.digest(formats[col].getSQLString(value).getBytes(StandardCharsets.UTF_8));
							else {
								nulls[col]++;
								currentHashes[col + 1] = empty;
							}
							if(consolidatedHashes[col+1]!=null)
								colMessageDigest.update(consolidatedHashes[col+1]);
							consolidatedHashes[col+1]=colMessageDigest.digest(currentHashes[col+1]);
							lineMessageDigest.update(currentHashes[col+1]);
						}
						currentHashes[0]=lineMessageDigest.digest();
						if(consolidatedHashes[0]!=null)
							lineMessageDigest.update(consolidatedHashes[0]);
						consolidatedHashes[0]=lineMessageDigest.digest(currentHashes[0]);
					}
					out.println("Columns");
					for(int col=0;col<formats.length;col++) {
						out.println(String.format("%1$4d",col+1)+". "+md.getColumnName(col+1));
						out.println("      - BaseType    : "+formats[col].name());
						out.println("      - SQL-Type    : "+md.getColumnTypeName(col+1));
						out.println("      - Values/Nulls: "+(rows-nulls[col])+"/"+nulls[col]);
						out.println("      - Hash        : "+toHex(consolidatedHashes[col+1]==null?empty:consolidatedHashes[col+1]));
					}

					out.println("Total");
					out.println("  - Records: "+rows);
					out.println("  - Hash   : "+toHex(consolidatedHashes[0]==null?empty:consolidatedHashes[0]));

					out.flush();
					out.close();
					if(tmpDir!=null) {
						File newFile=new File(destDir, filename);
						if (newFile.isFile()) {
							if (!newFile.delete())
								LOG.severe("Failed to delete target file "+newFile);
						}  else if(newFile.exists())
							LOG.severe("Target filename is not an accessible file: "+newFile);

						outfile.renameTo(new File(destDir, filename));
					}
				}
			}
		}
	}
	public static void main(String[] args) {
		main(SQLAnalyser.class, args);
	}
}
